/*
 * ----------------------------------------------------------------------------
 ghl - GartenHausLicht - Ein Mini-Solar-Projekt
 * ----------------------------------------------------------------------------
  rolf-harry 24.02.2007
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#define PULSE_MS (clock_pulse==0x01)    // das gibt 1 Impuls alle 131ms (1/1MHz/2**16*2)
#define PULSE_MS1 (clock_pulse==0x02)   // halb so oft

#define LED_RD_ON      (PORTB &= ~_BV(PB0))
#define LED_RD_OFF     (PORTB |= _BV(PB0))
#define LED_YELLOW_ON  (PORTA &= ~_BV(PA7))
#define LED_YELLOW_OFF (PORTA |= _BV(PA7))
#define LED_GREEN_ON   (PORTA &= ~_BV(PA5))
#define LED_GREEN_OFF  (PORTA |= _BV(PA5))
#define PANEL_SHORT    (PORTB |= _BV(PB1))
#define PANEL_ON       (PORTB &= ~_BV(PB1))
#define LAMP_ON        (PORTB |= _BV(PB2))
#define LAMP_OFF       (PORTB &= ~_BV(PB2))

//- globals...
volatile uint8_t clock;
volatile uint8_t clock_old;
volatile uint8_t clock_pulse;
uint16_t adc_ubat;  // Teiler 53,37
uint16_t adc_ldr;   //
uint16_t adc_pir;
uint16_t adc_temp;
uint16_t lamp_timer;
uint8_t flag_pir_off=0;
uint8_t rmp=0;
uint8_t rmp_step=0;

struct{
  uint8_t PWR_OFF:1;
  uint8_t PWR_LOW:1;
  uint8_t PWR_MED:1;
  uint8_t PWR_HIGH:1;
  uint8_t PWR_TOO_HIGH:1;
  uint8_t IS_DARK:1;
}flags;


//- functions...
//----------------------------------------------------------------------------
uint16_t ReadChannel(uint8_t);
void do_loop(void);
//-

//----------------------------------------------------------------------------
ISR (TIM1_OVF_vect)		
{
//  if(clock) clock=0;
//  else clock=1;
//  do_loop();
  clock++;
  
}
//----------------------------------------------------------------------------

ISR (TIM1_COMPA_vect)
{
  clock++;
}

//----------------------------------------------------------------------------
void do_loop(void){     // Die zyklische Schleife...

 
  // Die Analog-Kanle lesen, externe Referenz...
  adc_ubat=ReadChannel(1 | _BV(REFS0));
  adc_ldr=ReadChannel(2 | _BV(REFS0));
  adc_pir=ReadChannel(3 | _BV(REFS0));
  // Die Temperatur aber mit interner Referenz!
  adc_temp=ReadChannel(34 | _BV(REFS1));

// Batterie-Status ermitteln...
      if(adc_ubat<587){         //<11.0V Tiefentladung
        flags.PWR_OFF=1;
      }
      else if(adc_ubat>590){    //11.0V
        flags.PWR_OFF=0;
      }
        flags.PWR_LOW=0;
        flags.PWR_MED=0;
        flags.PWR_HIGH=0;
    
      if(adc_ubat>=590){        //11.0V schwach        
        flags.PWR_LOW=1;
        flags.PWR_MED=0;
        flags.PWR_HIGH=0;
      }
      if(adc_ubat>=640){        //12.0V mittel
        flags.PWR_LOW=0;
        flags.PWR_MED=1;
        flags.PWR_HIGH=0;
      }
      if(adc_ubat>=678){        //12.7V voll
        flags.PWR_LOW=0;
        flags.PWR_MED=0;
        flags.PWR_HIGH=1;
      }
     //14.0V Ladeendspannung bis 30GrC, ber 30GrC nur noch 13.2V
      if(((adc_ubat>=704) && (adc_temp>317)) || ((adc_ubat>=744) )){       
        flags.PWR_TOO_HIGH=1;
      }
      else if(adc_ubat<=690){ //13.0V
        flags.PWR_TOO_HIGH=0;
      }
  
// PIR - Bewegung erkennen, Lampe aber nicht ein, wenn Batterie zu schwach!
      if((adc_pir>820)&&flag_pir_off&&!rmp&&!flags.PWR_OFF&&flags.IS_DARK){    //.0V
        flag_pir_off=0;
        rmp=1;  // Lampen-Rampe-Start
      }
      else if(adc_pir<610){
        flag_pir_off=1;
      }

// LDR - Dunkelheit erkennen...
      if(adc_ldr>400){                          //1.0V
        flags.IS_DARK=1;
      }
      else if(adc_ldr<350){
        flags.IS_DARK=0;
      }

// Die LEDs nur bei Bewegung anzeigen...
// Bei Erreichen der Ladeendspannung gehen alle an
  LED_RD_OFF;
  LED_YELLOW_OFF;
  LED_GREEN_OFF;
  if((adc_pir>500)&&!flags.PWR_OFF){
    if(flags.PWR_LOW||flags.PWR_TOO_HIGH) LED_RD_ON;
    if(flags.PWR_MED||flags.PWR_TOO_HIGH) LED_YELLOW_ON;
    if(flags.PWR_HIGH||flags.PWR_TOO_HIGH) LED_GREEN_ON;
  }

// xxxxxxxxxxxxxxTEST der Temperatur-Messung
/*
LED_RD_OFF;
LED_YELLOW_OFF;
LED_GREEN_OFF;
if(adc_temp>300) LED_RD_ON;
if(adc_temp>305) LED_YELLOW_ON;
if(adc_temp>310) LED_GREEN_ON;
*/      
// xxxxxxxxxxxxxxTEST


// Bei Ladeendspannung das Panel kurzschliessen...
  if(flags.PWR_TOO_HIGH) PANEL_SHORT;
  else PANEL_ON;    
    
// Die Lampe hochdimmen, anlassen, und dann runterdimmen, und ausschalten...    
    switch(rmp){
      case(1):{         // PWM starten
        TCCR0A = _BV(COM0A1)|_BV(WGM00)|_BV(WGM01); 
        TCCR0B =  _BV(CS00);
        rmp_step = 0;
        rmp++;
      break;
      }
      case(2):{         // PWM++, dann Lampe an
        if(rmp_step<63)rmp_step++;
        else{
        TCCR0A = 0;
        TCCR0B = 0;
        lamp_timer=400; // ca. 25s
        LAMP_ON;  
        rmp++;      
        }
      break;
      }
      case(3):{         // PWM--
        if(lamp_timer)lamp_timer--;
        else{
          TCCR0A = _BV(COM0A1)|_BV(WGM00)|_BV(WGM01); 
          TCCR0B =  _BV(CS00);
          rmp++;        
        }
      break;
      }
      case(4):{         // PWM aus, Lampe aus
        if(rmp_step>1)rmp_step--;
        else{
          TCCR0A = 0;
          LAMP_OFF;
          rmp=0;
          flag_pir_off=1;
        }
      break;
      }
    } //switch(rmp)
    OCR0A=(rmp_step*4);  
  
}
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
void
ioinit (void)			/* Die Peripherie initialisieren */
{

TCCR1B =  _BV(CS10);          // Prescaler /1, ergibt 62ms bei 1MHz CLOCK 
TIMSK1 = _BV(TOIE1);          // enable T1 overflow interrupt 


/*
TCCR1B |= _BV(CS10);       // Prescaler 256
TCCR1A |= _BV(WGM12);     //Clear T1 on compare-match (CTC-mode)
OCR1A = 16383;
TIMSK1 |= _BV(OCIE1A);   //bei match int
*/
  sei ();
  DDRA =  _BV(DDA5);      // LED "C" grn
  DDRA |= _BV(DDA7);      // LED "A" gelb
  DDRB =  _BV(DDB0);      // LED "B" rot       
  DDRB |= _BV(DDB1);      // Panel Kurzschluss
  DDRB |= _BV(DDB2);      // Lampe
}
//----------------------------------------------------------------------------
//
//----------------------------------------------------------------------------
uint16_t ReadChannel(uint8_t mux)       // Einen Analog-Kanal lesen, von www.mikrocontroller.net
{
//-  uint8_t i;
  uint16_t result = 0;         //Initialisieren wichtig, da lokale Variablen
                               //nicht automatisch initialisiert werden und
                               //zufllige Werte haben. Sonst kann Quatsch rauskommen
  ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);    // Frequenzvorteiler 
                               // setzen auf 8 (1) und ADC aktivieren (1)
 
  ADMUX = mux;                      // Kanal waehlen
//  ADMUX |= (1<<REFS1) | (1<<REFS0); // interne Referenzspannung nutzen 
//-per Parameter...   ADMUX |= _BV(REFS0); // externe Referenzspannung nutzen 
  /* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
     also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
  ADCSRA |= (1<<ADSC);              // eine ADC-Wandlung 
  while ( ADCSRA & (1<<ADSC) ) {
     ;     // auf Abschluss der Konvertierung warten 
  }
 
  /* Eigentliche Messung - Mittelwert aus 4 aufeinanderfolgenden Wandlungen */
//-  for(i=0;i<4;i++)
//-  {
    ADCSRA |= (1<<ADSC);            // eine Wandlung "single conversion"
    while ( ADCSRA & (1<<ADSC) ) {
      ;   // auf Abschluss der Konvertierung warten
    }
//-    result += ADCW;		    // Wandlungsergebnisse aufaddieren
//-  }
  ADCSRA &= ~(1<<ADEN);             // ADC deaktivieren (2)
 
  result /= 4;                     // Summe durch vier teilen = arithm. Mittelwert
 
//-  return result;
  return ADCW;  //- > nur 1 Wandlung
}
//----------------------------------------------------------------------------
//
//----------------------------------------------------------------------------
int main (void)         // Haupt
{
  ioinit ();
  /* loop forever, the interrupts are doing the rest */
  for (;;){
    clock_pulse =  clock&(~clock_old);
    clock_old = clock; 
  
    do_loop();
    sleep_mode();
   }
    return (0);
}
//----------------------------------------------------------------------------
